:- multifile parameterization/2.
:- multifile parameter_values/2.
:- discontiguous parameterization/2.
:- discontiguous parameter_values/2.

%% portray(X)
%%	for syntactic objects
%%	abbrv(X) 	ellipsis used if nesting beyond threshold
%% parameters:
%% 	threshold(2)		default value 2
%% 	occ_debug		(off) prints @ if occ(V), var(V)
%% uses:
%%	boundaryMark(X)
%%

:- dynamic occ_debug/0.
parameterization(occ_debug,'use @ to indicate occ(V), V a variable').
% occ_debug :- fail.
% occ_debug.

%% printing
threshold(2).			% ellipsis used if nesting beyond threshold

portray(X) :- portray(X,_).
portray(set(L),Occ) :-		% FormSet
	format('{'),
	portray_list(L,Occ),
	format('}').
portray(pm([X,Y]),Occ) :-		% pair merge
    format('<'), portray(X,0,Occ), format(','), portray(Y,0,Occ), format('>').

portray(X,Occ) :- 
    (functor(X,so,5) ->
	 portray(X,0,Occ)
     ;   (X = abbrv(Y) ->
	      (boundaryMark(Y) -> format(b) ; portray(Y,1,_))
	  ;   portray(X,0,Occ))). % kinda wrong logic...

portray_list([],_Occ).
portray_list([X],Occ) :-
	!,  % red
	portray(X,0,Occ).
portray_list([X|L],Occ) :-
	portray(X,0,Occ),
	format(','), 
	portray_list(L,Occ).

mark_copies(X,Y) :- mark_copies(X,_,Y).
    
mark_copies(so(C,UFs,IFs,Fs,SOs),Occ1,so(Cp,UFs,IFs,Fs,SOsp)) :-
    propagateOcc(Occ1,Occ),
    mark_leC(C,Fs,SOs,Occ,C1), 
    mark_copiesC(C1,Fs,SOs,Occ,C2), 
    mark_copiesFs(UFs,'',X), 
    atom_concat(C2,X,Cp),
    mark_copiesL(SOs,Occ,SOsp).
mark_copies(pm([X,Y]),Occ1,pm([Xp,Yp])) :-
    propagateOcc(Occ1,Occ),
    mark_copies(X,Occ,Xp),
    mark_copies(Y,Occ,Yp).
mark_copies(set(List),Occ1,set(Listp)) :-
    propagateOcc(Occ1,Occ),
    mark_copiesL(List,Occ,Listp).

% prepending _ instructs tree drawing to put a line through
mark_copiesC(C,Fs,SOs,Occ,LBp) :-
	label_value(C,LB),
	(member(occ(X),Fs) ->
	 (var(X) ->
	      LBp = LB,
	      X = 1		% occ(1)
	  ;   (SOs == [] ->
		   atom_concat('_',LB,LBp)
	       ;  LBp = LB,
		  Occ = 1))
     ;  ((nonvar(Occ), SOs == []) -> % leaf: Occ set already
	     atom_concat('_',LB,LBp)
	 ;  LBp = LB)).

% first occurrence of phase(V), nonvar(V) gets "b_" prepended to label C
% printing: phases will be printed in blue
mark_leC(C,Fs,SOs,Occ,LBp) :-
	label_value(C,LB),
	(member(occ(X),Fs) ->
	 (var(X) ->
	  ((member(phase(V),Fs), nonvar(V)) ->
	   atom_concat('b_',LB,LBp)
	  ;   LBp = LB)
	 ;   LBp = LB)
	;  ((nonvar(Occ), SOs == []) -> % leaf: Occ set already
	     LBp = LB
	 ;   ((member(phase(V),Fs), nonvar(V)) ->
		  atom_concat('b_',LB,LBp)
	      ;   LBp = LB))).

% N:0, print everything
% N:1, print down to depth (threshold/1)
portray(inactive(X),N,Occ) :-		% inactive stack element
    !,					% red
    format('inactive('),
    portray(X,N,Occ),
    format(')').
portray(once(X),N,Occ) :-		% once stack element
    !,					% red
    format('once('),
    portray(X,N,Occ),
    format(')').
portray(set(L),N,Occ) :-	% formSet
	(N == 0 -> N1 = 0 ; N1 is N+1),
	format('{'),
	portray_list(L,N1,Occ),
	format('}').	
portray(pm([X,Y]),N,Occ) :-		% pair merge
    label_basetype(X,T), 
    !,					% red
    N1 is N+1,
    format('<'),
    (headInitial(T) ->		% <X,Y> when X is head-initial
	 portray(X,N1,Occ),
	 format(','),
	 portray(Y,N1,Occ)
     ;   portray(Y,N1,Occ),	% <Y,X> otherwise
	 format(','),
	 portray(X,N1,Occ)),
    format('>').
portray(so(C,UFs,_,Fs,SOs),N,Occ1) :-
	(N == 0 ->
	     propagateOcc(Occ1,Occ),
	     portrayC(C,Fs,SOs,Occ),
	     portrayFs(UFs),
	     portrayOcc(Fs),
	     portrayL(SOs,N,Occ), 
	     format(']')
	 ; (threshold(M),
	    label_value(C,LB),
	    (N>M ->
	     format('[~w..]',[LB])
	    ;
	     format('[~w',[LB]), 
	     portrayFs(UFs),
	     portrayOcc(Fs),
	     N1 is N+1,
	     portrayL(SOs,N1,Occ1), 
	     format(']')))).

portray_list([],_N,_Occ).
portray_list([X],N,Occ) :-
	!,  % red
	portray(X,N,Occ).
portray_list([X|L],N,Occ) :-
	portray(X,N,Occ),
	format(','), 
	portray_list(L,N,Occ).

portrayC(C,Fs,SOs,Occ) :-
	label_value(C,LB),
	((print_flag, member(occ(X),Fs)) ->
	 (var(X) ->
	  format('[~w',[LB]), 
	  X = 1			% occ(1)
	 ;
	  (SOs == [] ->
	   format('[_~w',[LB])
	  ;
	   format('[~w',[LB]), 
	   Occ = 1))
	;  ((nonvar(Occ), SOs == []) ->
	    format('[_~w',[LB])
	   ;
	    format('[~w',[LB]))).

% new default behavior: first copy of uF is printed, other copies omitted
portrayFs([]).
portrayFs([F|Fs]) :- 
    (F = f(FN,V) -> true ; F = f(FN,V,Occ)),
    ((unvalued(V), var(Occ)) -> format('!~w',[FN]), Occ = 1 ; true),
    portrayFs(Fs).

portrayOcc(Fs) :-
    ((occ_debug, member(occ(V),Fs),var(V)) ->
	 format('@~w',[V])
     ; true).

mark_copiesFs([],X,X).
mark_copiesFs([F|Fs],X,Y) :- 
    (F = f(FN,V) -> true ; F = f(FN,V,Occ)),
    ((unvalued(V), var(Occ)) ->
	 atom_concat('!',FN,FNp), Occ = 1, atom_concat(X,FNp,Xp)
     ;   Xp = X),
    mark_copiesFs(Fs,Xp,Y).

portrayL([],_N,_Occ).
portrayL([SO|SOs],N,Occ) :- portray(SO,N,Occ), portrayL(SOs,N,Occ).

mark_copiesL([],_Occ,[]).
mark_copiesL([SO|SOs],Occ,[SOp|SOsp]) :- 
    mark_copies(SO,Occ,SOp), mark_copiesL(SOs,Occ,SOsp).

% Occ will be Occ1 if Occ1 has been instantiated, otherwise Occ is fresh
propagateOcc(Occ1,Occ) :- nonvar(Occ1) -> Occ = Occ1 ; true.
